package ftpcomm;

use strict;
use warnings;

use Socket;
use Carp;
use File::Path;

my $EOL = "\015\012";
my $waitedpid = 0;

my $mod  = "HPCALE FTP Daemon";
my $lpipe = undef;
my $epipe = undef;

#forward declaration
sub REAPER;
sub handle_client;

sub set_env {
   ($lpipe, $epipe) = @_;
}

sub init_server{
   my ($host, $port, $proto);

   $host = $ENV{HPCALE_FTPD_HOST};
   $port = $ENV{HPCALE_FTPD_PORT};
   $proto= getprotobyname('tcp');

   # create socket
   socket(Server, PF_INET, SOCK_STREAM, $proto) 
   or die utilerr::prtvberr($mod,"001",$!);
   setsockopt(Server,SOL_SOCKET,SO_REUSEADDR, pack("l",1))
   or die utilerr::prtvberr($mod,"002",$!);

   # bind and listen
   bind(Server, sockaddr_in($port, INADDR_ANY)) or die utilerr::prtvberr($mod,"003",$!);
   listen(Server, SOMAXCONN) or die utilerr::prtvberr($mod,"004",$!);

   debug::logmsg($mod, "STDOUT", "Server is started on port $port");
   debug::logmsg($mod, $lpipe, "Server is started on port $port");


   my $paddr;
   $SIG{CHLD} = \&REAPER;

   for ($waitedpid = 0 ; 
        $paddr = accept(Client, Server) || $waitedpid ;
        $waitedpid = 0, close Client) {
 
      next if ($waitedpid and not $paddr);
      
      my($port,$iaddr) = sockaddr_in($paddr);
      my $name = gethostbyaddr($iaddr, AF_INET);

      debug::logmsg($mod, "STDOUT", "connection from $name ". inet_ntoa($iaddr) . " at port $port");
      debug::logmsg($mod, $lpipe, "connection from $name ". inet_ntoa($iaddr) . " at port $port");

      &handle_client;
   }
}

use POSIX ":sys_wait_h";
sub REAPER 
{
   my $child;
   while (($waitedpid = waitpid(-1,WNOHANG)) > 0) {
      debug::logmsg($mod, $lpipe, "reaped $waitedpid" . ($? ? " with exit $?" : ""));
   }
   $SIG{CHLD} = \&REAPER; 
}
   
sub handle_client 
{
   my $result = 0;

   # check certificate first
   if( check_cert() == 1){
      send(Client, '1', 0); # notify client certificate test

      my $jobid = &get_jobid();

      if(not defined $jobid){
         $result = 0;
      }
      else{
         debug::logmsg($mod, $lpipe, "transferring files...");
         my $type = recv_purpose();
         my $nfiles = recv_nfiles();

         if($type == 0){
            if($nfiles >= 0){ # if nfiles is 0, it means one file
               $result = relocate_files($jobid, $nfiles);
            }

            if($result == 1){
               debug::logmsg($mod, $lpipe, "relocating files: ok");
            }
            else{
               debug::logmsg($mod, $lpipe, "relocating files: fail");
            }
         }
         else{
            if($nfiles >= 0){
               $result = relocate_back_files($jobid, $nfiles);
            }

            if($result == 1){
               debug::logmsg($mod, $lpipe, "relocating back files: ok");
            }
            else{
               debug::logmsg($mod, $lpipe, "relocating back files: fail");
            }
         }
      }
   }
   
}

sub relocate_back_files 
{
   my ($jobid, $nfiles) = @_;

   my $size = '';
   my $srchost = '';
   my $filename = '';
   my $fnprefix = "$ENV{HPCALE_JOBDIR}/$jobid/tempresult";
   
   # make temporary directory to hold result files
   mkpath(["$fnprefix"], 0, 0711);

   foreach (0..$nfiles){
      # recv src host and filename
      $srchost = &recv_str();
      $filename = &recv_str();

      debug::dbgmsg(3, $mod, $lpipe, "$srchost:$filename");

      my $r = &recv_from_rc($jobid, $srchost, $filename, $fnprefix);
      if($r == 1){
         debug::logmsg($mod, $lpipe, "grabbing $filename: ok");
         send(Client, '1', 0); # notify client the result of the file transfer
      }
      else{
         debug::errmsg($mod, $lpipe, "grabbing $filename: failed");
         send(Client, '0', 0); # notify client the result of the file transfer
         return 0;
      }

      $r = &send2client($fnprefix, $filename);
      if($r == 1){
         debug::logmsg($mod, $lpipe, "back-transfer $filename: ok");
         send(Client, '1', 0); # notify client the result of the file relocation
      }
      else{
         debug::errmsg($mod, $lpipe, "back-transfer $filename: failed");
         send(Client, '0', 0); # notify client the result of the file relocation
         return 0;
      }
   }

   return 1;
}

#
# assume that every output files are not grabbed into 
# the $fnprefix directory of the server (directory structure of the output files
# are lost at server side and may be restructured at client side as specified in
# the recvfiles tag in XJD
#
sub send2client 
{
   my ($fnprefix, $filename) = @_;
   my $data = '';
   
   if($filename =~ /(.*)\/(.+)/){
      $filename = $2;
   }

   debug::dbgmsg(3, $mod, $lpipe, "File: $fnprefix/$filename");
   open TOSEND, "< $fnprefix/$filename";
   binmode TOSEND;

   my $size = -s TOSEND;
   &send_int($size);
   debug::dbgmsg(3, $mod, $lpipe, "size: $size");

   # no need to send
   if($size == 0){
      close TOSEND;
      return 1;
   }

   # send file contents
   my $chunksize = 1000;
   my $quot = $size / $chunksize;
   my $remain = $size % $chunksize;

   my $i = 1;
   my $sendok = 1;
   while($i <= $quot) {
      if($sendok == 1){
         read(TOSEND, $data, $chunksize);
      }

      my $realsent = send(Client, $data, 0);

      recv(Client, $data, 1, 0);
      if($data == 1){
         debug::dbgmsg(3, $mod, $lpipe, "send ok");
         $sendok = 1;
         $data = '';
      }
      else{
         debug::dbgmsg(3, $mod, $lpipe, "send fail.  try again.");
         $sendok = 0;
      }
      $i++;
   }

   $data = '';
   my $a = 0;
   while(1){
      read(TOSEND, $data, $remain) if (length($data) == 0);

      send(Client, $data, 0);
      recv(Client, $a, 1, 0);

      last if($a == 1);
   }

   # end of file contents transfer
   close TOSEND;
   return 1;
}

sub recv_from_rc
{
   my ($jobid, $srchost, $filename, $fnprefix) = @_;

   my $r = undef;
   my $srcfp = undef;
   my $srcfn = undef;

   if($filename =~ /(.*)\/(.+)/){
      $srcfp = $1;
      $srcfn = $2;
   }
   else{
      $srcfp = "";
      $srcfn = $filename;
   }

   &gen_grab_batchfile($fnprefix, $srcfp, $srcfn, $jobid);

   # return code seems to be 0 for successful execution. otherwise, 256
   $r = system ("sftp -b $fnprefix/batch-$srcfn $srchost");
   ($r == 0) ? return 1 : return -1;
}

sub gen_grab_batchfile 
{
   my ($fnprefix, $srcfp, $srcfn, $jobid) = @_;
   my $srcdir = undef;

   unlink "$fnprefix/batch-$srcfn";

   if($srcfp =~ /^\//){ # absolute path
      $srcdir = $srcfp;
   }
   else{ # relative path
      $srcdir = "$ENV{HPCALE_REMOTE_JOBDIR}/$jobid/$srcfp";
   }

   open BATCH, "> $fnprefix/batch-$srcfn";
   print BATCH "lcd $fnprefix\n";
   print BATCH "get $srcdir/$srcfn\n";
   close BATCH;
}

sub relocate_files 
{
   my ($jobid, $nfiles) = @_;

   debug::dbgmsg(3, $mod, $lpipe, "jobid: $jobid, #file: $nfiles (+1)");
   
   my $desthost = '';
   my $filename = '';
   my $fnprefix = "$ENV{HPCALE_JOBDIR}/$jobid";

   foreach (0 .. $nfiles){
      # recv destination host
      $desthost = &recv_str();
      $filename = &recv_str();
      debug::dbgmsg(3, $mod, $lpipe, "desthost:$desthost, file:$filename");

      # decide temporary file name at HPCALE server
      $filename =~ /.*\/(.+)/;
      my $fn = $1;
      my $tmpfn = "$fnprefix/$fn";

      debug::dbgmsg(3, $mod, $lpipe, "recv to: <$tmpfn>");

      # recv file 
      my $r = recv_file ($tmpfn);
      if($r != 1){
         debug::errmsg($mod, $lpipe, "receiving $filename from client: failed");
         send(Client, '0', 0); # notify client the result of the file transfer
         return 0;
      }

      $r = &send2rc($jobid, $desthost, $fnprefix, $filename, $fn);
      if($r == 1){
         debug::logmsg($mod, $lpipe, "relocating $filename: ok");
         send(Client, '1', 0); # notify client the result of the file relocation
      }
      else{
         debug::errmsg($mod, $lpipe, "relocating $filename: failed");
         send(Client, '0', 0); # notify client the result of the file relocation
         return 0;
      }
   }

   return 1;
}

sub send2rc
{
   my ($jobid, $desthost, $fnprefix, $filename, $fn) = @_;
   debug::logmsg($mod, $lpipe, "sending $fnprefix/$fn to $desthost");

   my $r = undef;
   my $tmpfn = "$fnprefix/$fn";

   $filename =~ /(.*)\/(.+)/;
   my $dstfp = $1;
   my $dstfn = $2;

   $r = dirman::create_remote_dir($desthost, $dstfp, $jobid);
   return -1 if($r < 0);

   &gen_batchfile($fnprefix, $fn, $dstfp, $dstfn, $jobid);

   # return code seems to be 0 for successful execution. otherwise, 256
   $r = system ("sftp -b $fnprefix/batch-$dstfn $desthost");
   ($r == 0) ? return 1 : return -1;
}

sub gen_batchfile 
{
   my ($fnprefix, $fn, $dstfp, $dstfn, $jobid) = @_;
   my $dstdir = undef;

   unlink "$fnprefix/batch-$dstfn";

   if($dstfp =~ /^\//){ # absolute path
      $dstdir = $dstfp;
   }
   else{ # relative path
      $dstdir = "$ENV{HPCALE_REMOTE_JOBDIR}/$jobid/$dstfp";
   }

   open BATCH, "> $fnprefix/batch-$dstfn";
   print BATCH "cd $dstdir\n";
   print BATCH "put $fnprefix/$fn\n";
   close BATCH;
}

sub recv_file 
{
   my $filename = shift;

   my $size = recv_int(); # recv file size
   my $data = '';

   debug::dbgmsg(3, $mod, $lpipe, "<<$filename, $size>>");

   if(not defined $size || length($size) == 0){
      debug::errmsg($mod, $lpipe, "file size is 0.");
      send(Client, "0", 0);
      return 0;
   }
   if(not defined $filename || length($filename) == 0){
      debug::errmsg($mod, $lpipe, "file name is not specified.");
      send(Client, "0", 0);
      return 0;
   }

   # send ok-to-send signal to client
   send(Client, "1", 0);


   my $chunksize = 1000;
   my $quot = $size / $chunksize;
   my $remain = $size % $chunksize;

   # start recv
   open TORECV, "> $filename";
   binmode TORECV;

   my $i = 1;
   while ($i <= $quot){
      recv(Client, $data, $chunksize, 0);

      if(length($data) < 1000){
         print "data recvd: " . length($data) . "\n";
         send(Client, "0", 0);
         next;
      }
      else{
         send(Client, "1", 0);
      }

      print TORECV $data;
      $data = '';
      $i++;
   }

   $data = '';
   recv(Client, $data, $remain, 0);
   print TORECV $data;

   close TORECV;
   # end recv

   my $recvdsize = -s $filename;
   ($size == $recvdsize) ? return 1 : return 0;
}

sub recv_str
{
   my $str = '';
   my $size = recv_int(); # recv string length 
   recv(Client, $str, $size, 0);

   return $str;
}

sub recv_int 
{
   my $char = '';
   my $size = '';

   while(1) {
      recv(Client, $char, 1, 0);
      if($char eq '.'){
         last;
      }
      $size .= $char;
   }

   return $size;
}

sub send_int
{
   my $int = shift;
   send(Client, "$int.", 0); # send integer and end mark

   return 1;
}

sub check_cert 
{
   my $certhost = '';
   my $certfile = '';
   my $fnprefix = $ENV{HPCALE_FTPD_HOME};

   # recv certificate destination host name (dummy)
   $certhost = &recv_str();
   debug::logmsg($mod, $lpipe, "cert dest host: <$certhost>");
   
   # recv certificate file name
   $certfile = &recv_str();
   $certfile = "$fnprefix/$certfile";
   debug::logmsg($mod, $lpipe, "filename: <$certfile>");

   # recv actual certificate 
   recv_file ($certfile);

   # notify client that file is received
   send(Client, '1', 0); 

   # check certificate sanity 
   my $openssl = $ENV{OPENSSL_BIN};
   my $capath  = $ENV{HPCALE_FTPD_CADIR};
   my $cafile  = $ENV{HPCALE_FTPD_CAFILE};
   my $cert    = $certfile;

   my $result = `$openssl verify -CApath $capath -CAfile $cafile $certfile`;
   chomp $result;
   if ($result =~ /OK/){
      debug::logmsg($mod, $lpipe, "Certificate check passed: <$result>");
      return 1;
   }
   else{
      debug::logmsg($mod, $lpipe, "unacceptable certificate");
      return 0;
   }
}

sub get_jobid
{
   my $jobid = recv_int();

   my $fnprefix = $ENV{HPCALE_JOBDIR};
   if(-e "$ENV{HPCALE_JOBDIR}/$jobid"){
      return $jobid;
   }
   else{
      return undef;
   }
}

sub recv_purpose
{
   # recv the purpose
   debug::logmsg($mod, $lpipe, "receiving purpose");
   my $type = recv_int();
   if($type == 0){
      debug::logmsg($mod, $lpipe, "connection to transfer input files");
   }
   else{
      debug::logmsg($mod, $lpipe, "connection to transfer output files");
   }

   return $type;
}

sub recv_nfiles
{
   # recv the number of files
   my $nfiles = recv_int();
   if($nfiles < 0){
      debug::logmsg($mod, $lpipe, "nothing to transfer.");
   }
   else{
      my $t = $nfiles + 1;
      debug::logmsg($mod, $lpipe, "must transfer $t files.");
   }

   return $nfiles;
}

1;
