# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package RCAllocate;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}");
    push (@INC, "$ENV{HPCALE_SERVER_HOME}/server");
}

use allocation::RC_Analyzer;
use allocation::RC_Locker;
use allocation::ResourceReserver;
use util::XJDParser;
use util::Debug;
use util::RemoteDirManager;
use util::JobRecordManager;
use File::Path;
use strict;

my $ALLOCTEST = 0;
my $mod = "RCAllocate";
my $pipe= "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";


sub new {
   my $classname  = shift;         # What class are we constructing?
   my $self   = {};                # Allocate new memory

   bless($self, $classname);       # Mark it of the right type

   return $self;
}

sub allocateResource {
   my ($self,$user,$host,$lowXJD) = @_;

   my $jobRecordMgr  = new JobRecordManager();
   my $locker        = new RC_Locker();
   my $reserver      = new ResourceReserver();
   
   my $jobid = $jobRecordMgr->assign_job_id($user, $host);
   if($jobid < 0){
       Debug::logmsg($mod, $pipe, "Job id cannot be assigned");
       return ($jobid, -1);    
   }

   my $jobDir = "$ENV{HPCALE_JOBDIR}/$jobid";
   my $WaitScriptLoc = "$ENV{HPCALE_SERVER_HOME}/server/allocation";
   my $PVM_startnode = undef;

   ############################################################
   # Step 1 : create job directories on resource managers (head)
   ############################################################
   my $createOk = &createJobDirectory($jobid,
                                      $jobDir,
                                      $lowXJD,
                                      $jobRecordMgr);
   if($createOk != 1){
       Debug::logmsg($mod, $pipe, "Job directory creation failed");
       return ($jobid, -1);    
   }

   ############################################################
   # Step 2 : organize the resource into files according to the 
   #          type of the resources
   ############################################################
   my @RCFileSet = ();
   my $rcOk = (new RC_Analyzer())->organizeResources($lowXJD, 
                                                     $jobDir, 
                                                     \@RCFileSet);
   if($rcOk == -1){
       Debug::logmsg($mod, $pipe, "Error in resource organizing for 
job $jobid...");       return ($jobid, -1);
   }

   my $numHosts = &TotalAskedHostNum(\@RCFileSet, $jobDir);
   Debug::dbgmsg(3, $mod, $pipe, "Total number of hosts : $numHosts");


   ############################################################
   # Step 3 : create locks in remote resource managers if needed
   #
   # Should create lock file since the script use this lock
   # file for staying without exiting.
   # The lock files will be removed by execution service after
   # the application terminates.
   ############################################################
   my $lockCreated = $locker->createRemoteLocks(\@RCFileSet, 
                                                $jobDir, 
                                                $jobid);
   if($lockCreated < 0){
       Debug::logmsg($mod, $pipe, "Remote Lock creation failed.");
       return ($jobid, -1);
   }                                                 
        

   ############################################################
   # Step 4 : reserve node
   # If successful, it means that all SSH-albe nodes are accessible,
   # and all node requests are sent to the schedulers and we got
   # the id assigned by schedulers. 
   ############################################################
   my ($reserveOk, @SchedReserved) = 
      $reserver->reserveNodes(\@RCFileSet,$jobDir,$jobid,$WaitScriptLoc);   

   if($reserveOk < 0){
      Debug::logmsg($mod, $pipe, "Reservation failed.");
      return ($jobid, -1);
   }
   
   
   ############################################################
   # Step 5 : wait until all reservations are satisfied 
   #  if a pbs server does not give resource within the limit, 
   #  HPCALE reports reservation failure.
   ############################################################
   my $status = $reserver->checkReserveStatus ($jobDir, 
                                               $locker, 
                                               $jobRecordMgr, 
                                               \@SchedReserved);
   if($status < 0){
      Debug::logmsg($mod, $pipe, "Reservation failed. Releasing resources...");
      $reserver->releaseResources(\@SchedReserved);
      return ($jobid, -1);
    }
    else{    
      $jobRecordMgr->update_record($jobid, 1, "allocated");
      $jobRecordMgr->update_record($jobid, 2, "pending");

      if($ALLOCTEST == 1){
         Debug::logmsg($mod, $pipe, "Releasing Resource... (Only for Allocation Test)");        
         $reserver->releaseResources(\@SchedReserved);         
         return ($jobid, 1);      
      }
   }
    
   return ($jobid, 1);
}


sub TotalAskedHostNum {
   my $RCFileSet = shift;
   my $jobDir = shift;

   my $num = 0;

   foreach(@{$RCFileSet}){
      my $isCluster = 0;

      open RCFILE, "< $jobDir/$_";

      my @arr = <RCFILE>;
      foreach(@arr){
         if( $_ =~ /.+,(.+)/ ){
            $isCluster = 1;
            $num += $1;
         }
      }

      close RCFILE;

      if($isCluster == 0){
         $num += scalar (@arr);
      }
   }

   return $num;
}

sub createJobDirectory {
   my ($jobid, $jobdir, $lowXJD, $jobRecordMgr) = @_;

   Debug::dbgmsg(3, $mod, $pipe, "Creating Job Directories...");
   Debug::dbgmsg(3, $mod, $pipe, "JobID : $jobid");  
   Debug::dbgmsg(3, $mod, $pipe, "JobDir: $jobdir");  

   # create working directory at server side
   # we can sure this directory is unique and didn't exist before
   # because the job id is unique.
   
   if(-e $jobdir){
      Debug::errmsg($mod, $pipe, "Directory $jobdir is already exist.");   
   }
   else{
      mkpath(["$jobdir"], 0, 0711);
   }

   my $dirmanager = new RemoteDirManager();
   my $headJobDir = $ENV{HPCALE_REMOTE_JOBDIR};

   # create working directory at each clusterhead
   my @hosts = ();
   my $compList = $lowXJD->getComponentList();

   foreach my $key (keys %{$compList}){
      my $comp = $compList->{$key};
      my $head = $comp->{head};

      my $duplicate = 0;
      foreach (@hosts){
         if($_ eq $head){
             $duplicate = 1;
         }
      }

      my $i = 0;
      if($duplicate == 0){
         push @hosts, $head;

         $i = $dirmanager->createRemoteDir($head, 
"$headJobDir/$jobid");         if($i < 0){
            my $str = "Remote directory for job $jobid cannot be 
created.";            Debug::errmsg($mod, $pipe, $str);
            return -1;
         }
         else{
            my $str = "Remote directory for job $jobid is created.";
            Debug::logmsg($mod, $pipe, $str);
         }
      }
   }

   $jobRecordMgr->update_record($jobid, 1, "pending");

   return 1;
}

1;
