# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package RC_Analyzer;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}");
}

use util::Debug;
use strict;

my $mod = "RC_Analyzer";
my $pipe= "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";


sub new {
   my $classname  = shift;         # What class are we constructing?
   my $self   = {};                # Allocate new memory

   bless($self, $classname);       # Mark it of the right type

   $self->{NodeRCfile} = "NodeRC";
   $self->{PBS_RCfilePostfix} = "PBSrc";
    
   return $self; 
}


sub organizeResources {
   my $self  = shift;
   my $lowXJD = shift;
   my $wdir  = shift;
   my $RCFileSet = shift;

   my $cList = $lowXJD->getComponentList();
   my $rcOk = 0;

   foreach my $key (keys %{$cList}){
      my $comp = $cList->{$key};
     
      my $str = join ",", "Key: <$key>", 
                          "Cluster: <$comp->{cluster}>"; 
      Debug::dbgmsg(3, $mod, $pipe, $str);
        
      # try to create rc file. if it's new rc file, insert into RCFileSet
      # we regard this case as running on a stand-alone node
      if(length($comp->{cluster}) == 0){
         Debug::dbgmsg(3, $mod, $pipe, "Federate_NodeRC");
         $rcOk = $self->federate_NodeRC($comp, $wdir, $RCFileSet);
      }
      else{
         Debug::dbgmsg(3, $mod, $pipe, "Federate_ClusterRC");
         $rcOk = $self->federate_ClusterRC($comp, $wdir, $RCFileSet);        
      }
        
      if ($rcOk < 0){
         return -1;
      }
   }
    
   return 1;
}


# Assume that 'Node' can be accessible by SSH directly.
# Write node into the file 'NodeRC' (allow duplicate)
sub federate_NodeRC($$$){
   my $self   = shift;
    
   my $comp   = shift;
   my $wdir   = shift;
   my $RCFileSet  = shift;
    
   my $head = $comp->{head};
   my $canonicalName = "$comp->{node}.$comp->{domain}";
    
   if(-e "$wdir/$self->{NodeRCfile}"){
      open RCFILE, ">> $wdir/$self->{NodeRCfile}" or return -2;
   }
   else{
      push @{$RCFileSet}, $self->{NodeRCfile};
      open RCFILE, "> $wdir/$self->{NodeRCfile}" or return -2;
   }

   print RCFILE "$canonicalName\n";
   close RCFILE;    
    
   return 1;
}


sub federate_ClusterRC($$$){
   my $self = shift;
    
   my $comp = shift;
   my $wdir = shift;
   my $RCFileSet = shift;
    
   my $rcOk = 1;
    
   if($comp->{scheduler} eq "PBS"){
      Debug::dbgmsg(3, $mod, $pipe, "Federate_PBS_ClusterRC");
      $rcOk = $self->federate_PBS_ClusterRC($comp, $wdir, $RCFileSet);
   }
   elsif($comp->{scheduler} eq "SSH"){
      Debug::dbgmsg(3, $mod, $pipe, "Federate_SSH_ClusterRC");
      $rcOk = $self->federate_SSH_ClusterRC($comp, $wdir, $RCFileSet);
   }
   
   return $rcOk;
}


sub federate_PBS_ClusterRC($$){
   my $self = shift;
   
   my $comp = shift;
   my $wdir = shift;
   my $RCFileSet = shift;
   
   my $RCFileName = "$comp->{cluster}.$self->{PBS_RCfilePostfix}";
    
   if(-e "$wdir/$RCFileName"){
      open RCFILE, ">> $wdir/$RCFileName" or return -2;
   }
   else{
      push @{$RCFileSet}, $RCFileName;
      open RCFILE, "> $wdir/$RCFileName" or return -2;
   }

   if(length($comp->{nodeset}) == 0){
      print RCFILE "$comp->{head},$comp->{nNode}\n";
   }
   else{
      # In principle, we can specify the specific nodes in the cluster 
      # for applications However, PBS seems to have problem in allocating 
      # nodes if '-l nodes=' is like follows. '-l nodes = 4+red10+red11+8'
      # Like this, mixed expression for nodes seems to be mistreated in 
      # the PBS. Thus, at this time, specifying exact node names inside 
      # the PBS cluster is treated as error.

      return -1;
    }
    
    close RCFILE;
    
    return 0;
}


sub federate_SSH_ClusterRC(){
   my $self = shift;
    
   my $comp = shift;
   my $wdir = shift;
   my $RCFileSet = shift;

   my $RCFileName = $self->{NodeRCfile};

   if(-e "$wdir/$RCFileName"){
      open RCFILE, ">> $wdir/$RCFileName" or return -2;
   }
   else{
      push @{$RCFileSet}, $RCFileName;
      open RCFILE, "> $wdir/$RCFileName" or return -2;
   }

   if(length($comp->{nodeset}) == 0){
      # In principle, separate daemon should find available nodes 
      # in this cluster and return them. Currently, treated as error.
      # This means that user must specify exact node names to use 
      # the nodes under SSH cluster.
        
      return -1;
    }
    else{
      my @nodes = split /,/ , $comp->{nodeset};  
      my $domain= &getClusterDomain($comp->{head});
            
      if(scalar(@nodes) != $comp->{nNode}){
         close RCFILE;

         my $str = " Number mismatch between nNode and nodeset ----> " . 
                   scalar(@nodes) .
                   " $comp->{nNode}";
         Debug::dbgmsg(3, $mod, $pipe, $str);

         return -1;
      }
        
      foreach(@nodes){
         print RCFILE "$_.$domain\n";
      }
   }
    
   close RCFILE;
    
   return 1;
}

sub getClusterDomain($){
   my $cluster = shift;

   $cluster =~ /[a-zA-Z0-9]+\.(.*)/;
   return $1;
}
   

1;

