# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package ResourceReserver;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}");
}

use util::Debug;
use util::RemoteDirManager;

use strict;

my $mod = "ResourceReserver";
my $pipe= "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";


sub new {
   my $classname  = shift;         # What class are we constructing?
   my $self   = {};                # Allocate new memory

   bless($self, $classname);       # Mark it of the right type

   $self->{NodeRCfile} = "NodeRC";
   $self->{PBS_RCfilePostfix} = "PBSrc";
   $self->{WaitScript} = "holdrc.pl";
   $self->{LockFile} = "hpcale_lock";
   $self->{jobid} = "";
   $self->{RESERVE_LIMIT} = $ENV{HPCALE_RESERVE_RETRY};

   return $self;
}

#
# reserve resources
#  - if all succeeds, return (1, schedreserved)
#  - else, return (-1, ())
#
sub reserveNodes{
   my $self = shift;
   my $RCFileSet = shift;
   my $serverDir = shift;
   my $jobid = shift;
   my $WaitScriptLoc = shift;

   my @SchedReserved = ();

   $self->{jobid} = $jobid;

   my $remoteJobdir = "$ENV{HPCALE_REMOTE_JOBDIR}/$jobid";
    
   foreach(@{$RCFileSet}){
      my $RC_Filefullname = "$serverDir/$_";

      if($_ =~ /$self->{NodeRCfile}/){
         my $nodeOk = $self->checkSSHableNodeAvailability ($RC_Filefullname, $remoteJobdir);
         if($nodeOk < 0){
            Debug::errmsg($mod, $pipe, "Some SSH nodes are not available. Resource reservation failed.");
            $self->releaseResources(\@SchedReserved);
            return (-1, ());
         }
      }
      elsif($_ =~ /$self->{PBS_RCfilePostfix}/){
         my $head = undef;
         my $PBSjobid = undef;
         my $nNode = undef;
         my $schedType = undef;
         
         ($head, $PBSjobid, $nNode, $schedType) = $self->reservePBSNodes($RC_Filefullname, $remoteJobdir, $WaitScriptLoc);

         if($PBSjobid < 0){
            Debug::errmsg($mod, $pipe, "Cluster managed by $head failed to return PBS job id.");
            $self->releaseResources(\@SchedReserved);
            return (-1, ());
         }
         else{
            Debug::dbgmsg(3, $mod, $pipe, "pushing [$head, $PBSjobid, $nNode, $schedType, 0]");
            push @SchedReserved, [$head, $PBSjobid, $nNode, $schedType, 0];
         }
      }
   }
    
   return (1, @SchedReserved);
}


sub checkSSHableNodeAvailability {
   my $self = shift;
   my $NodeRC_fullname = shift;
   my $jobDir = shift;
    
   open SSHNODES, "< $NodeRC_fullname" ;
   my @nodes = <SSHNODES>;
   close SSHNODES;
    
   foreach(@nodes){            
      chomp $_;
      my $nodeOk = $self->checkHostAvailability ($_, $jobDir);
        
      if($nodeOk < 0){
         Debug::errmsg($mod, $pipe, "Resource $_ is not available");
         return -1;
      }
   }
    
   return 1;
}

sub checkHostAvailability {
   my $self = shift;
   my $node = shift;
   my $path = shift;
   
   my $dirmanager = new RemoteDirManager();
   my $hostAlive = $dirmanager->isHostAlive($node);
   if($hostAlive < 0){
      Debug::errmsg($mod, $pipe, "Resource $_ seems to be dead");
      return -1;
   }
    
   my $home = $dirmanager->getHomeDir($node);
   my $direxist = $dirmanager->dirExist($node, "$home/$path");    
   if($direxist < 0){
      Debug::errmsg($mod, $pipe, "Job directory ($home/$path) error on resource $_.");
      return -1;
   }
   
   return 1;
}


sub reservePBSNodes {
   my $self = shift;
   my $PBSRC_fullname = shift;
   my $remotejobDir = shift; # means remote job dir
   my $WaitScriptLoc = shift;

   my $wait_script = $self->{WaitScript};
   my $LockFile = $self->{LockFile};    
    
   open PBSRC, "< $PBSRC_fullname";
   my @nNodes = <PBSRC>;
   close PBSRC;

   # Sum the number of asked nodes
   my $head;
   my $sumNodes = 0;
   foreach(@nNodes){
      $_ =~ /(.+),([0-9]+)/;
   
      Debug::dbgmsg(3, $mod, $pipe, "Asked Nodes: >>>$1 : $2 <<<");
        
      $head = $1;
      $sumNodes += $2;
   }
  

   # gather information on the remote cluster 
   my $dirManager = new RemoteDirManager();
   my $headhome = $dirManager->getHomeDir($head);
   my $qsubpath = $dirManager->checkProgramDir($head, "qsub"); #chomp $qsubpath;
   my $perlpath = $dirManager->checkProgramDir($head, "perl"); #chomp $perlpath;

   # add the first line of the wait script (depends on the perl location of cluster head)
   `echo '#!$perlpath -w' | cat - $WaitScriptLoc/$wait_script > $WaitScriptLoc/$wait_script.tmp`;
    
   # copy wait script to the cluster head and remove temporary wait script
   `cat $WaitScriptLoc/$wait_script.tmp | ssh $head 'cat - >! $headhome/$remotejobDir/$wait_script'`;
   `ssh $head 'chmod 700 $headhome/$remotejobDir/$wait_script'`;
   `rm $WaitScriptLoc/$wait_script.tmp`;

   my $qsubcmd = "ssh $head 'echo \"$headhome/$remotejobDir/$wait_script $LockFile $headhome/$remotejobDir\" | $qsubpath -l nodes=$sumNodes'";
   Debug::dbgmsg(3, $mod, $pipe, "QSUB Command: $qsubcmd");

   my $pbsjid = `$qsubcmd`;
   if($pbsjid =~ /([0-9]+)\..+/){
      return ($head, $1, $sumNodes, "PBS");
   }
   else{
      return ($head, -1, $sumNodes, "PBS");
   }
} 
    

sub releaseResources {
   my $self = shift;
   my $SchedReserved = shift;
   my $dirManager = new RemoteDirManager();

   Debug::logmsg($mod, $pipe, "Releasing resources...\n<<@{$SchedReserved}>>");

   foreach(@{$SchedReserved}){
      my $head = $_->[0];
      my $id = $_->[1];
      my $sched = $_->[3];

      my $qdelpath = $dirManager->checkProgramDir($head, "qdel"); 
      chomp $qdelpath;

      my @reply = ();
      if($sched eq "PBS"){
         Debug::logmsg($mod, $pipe, "Releasing $sched cluster managed by $head.");

         my $reply = `ssh $head '$qdelpath $id'`;
         if(length($reply) == 0){
            Debug::logmsg($mod, $pipe, "Qdel for job $id on $head is successful.");
         }
         else{
            Debug::errmsg($mod, $pipe, "Qdel Error: $reply.");
         }
      }
   }        
}

sub checkReservedNodeHealth {
   my $self = shift;
}

sub checkReserveStatus {
   my $self = shift;
   my $serverDir = shift;
   my $locker = shift;
   my $jManager = shift;
   my $SchedReserved = shift;
    
   my $status = undef;

   # make this into dbgmsg
   foreach(@{$SchedReserved}){
      Debug::dbgmsg(3, $mod, $pipe, "Head: $_->[0], PBSid: $_->[1], nNode: $_->[2], Sched: $_->[3]");
   }

   my $dirManager = new RemoteDirManager();

   my $allDone = 0;        
   for (my $retry = $self->{RESERVE_LIMIT}; $retry > 0 && $allDone == 0; $retry--){
      system("sleep 2");
      
      foreach my $cluster (@{$SchedReserved}){
         my $head = $cluster->[0];
         my $id = $cluster->[1];
         my $sched = $cluster->[3];
         my $isReserved = $cluster->[4];

         my $qstatpath = $dirManager->checkProgramDir($head, "qstat"); 

         if($isReserved == 0){
            if($sched eq "PBS"){
               my @reply = `ssh $head '$qstatpath -f $id | grep job_state'`;
               foreach(@reply){
                  if($_ =~ /job_state = ([A-Za-z])/){
                     Debug::dbgmsg(3, $mod, $pipe, "Job Status : $1 , $head");

                     if($1 eq "R"){  # reserved 
                        $cluster->[4] = 1;

                        my $headhome = (new RemoteDirManager())->getHomeDir($head);
                        my $headJobDir = "$headhome/$ENV{HPCALE_REMOTE_JOBDIR}/$self->{jobid}";
                               
                        # the PBSNODES file under $headJobDir is created by WaitScript(holdrc.pl) 
                        # by referring the $ENV{PBS_NODEFILE} set by PBS itself.
                        if((new RemoteDirManager())->fileExist($head, "$headJobDir/PBSNODES") == 1){
                           Debug::dbgmsg(3, $mod, $pipe, "Writing PBSNODE-$head file under $serverDir");

                           open PBSNODE, "> $serverDir/PBSNODES-$head";
                           my @nodes = `ssh $head 'cat $headJobDir/PBSNODES'`;
                           foreach(@nodes){
                              Debug::dbgmsg(3, $mod, $pipe, ">>>node:$_<<<");
                              print PBSNODE "$_";
                           }
                           close PBSNODE;                                
                        }
                        else{ #PBSNODE file is not ready 
                           $cluster->[4] = 0;
                        }
                     }
                  }
               }
            }
         }
      }

      $allDone = 1;
      foreach my $cluster (@{$SchedReserved}){
         my $isReserved = $cluster->[4];
         if($isReserved == 0){
            $allDone = 0;
            last;
         }
      }

      last if ($allDone == 1);
   }

   if($allDone == 1){
      Debug::logmsg($mod, $pipe, "Congrat! All resources are successfully reserved");
      return 1;
   }
   else{
      foreach my $cluster (@{$SchedReserved}){
         my $head = $cluster->[0];
         my $isReserved = $cluster->[4];
            
         if($isReserved == 0){
            Debug::errmsg($mod, $pipe, "Sorry. $head can't be reserved right now");
         }
      }
        
      $self->releaseResources($SchedReserved);
      return -1;
   }
}



##############################################################################
# BELOW, DEPRECATED FUNCTIONS
##############################################################################


# PBS node reservation ver.2 (deprecated)
# This function uses PBSNODE file generated by PBS scheduler
# But, use $mainleader, not $leader assigned for each component.
# Each component should use the leader assigned for them.
#
# reserve pbs-scheduled cluster nodes using "pvmaddatonce.pl"
# at one reserved host by checking PBS-reserved nodes
#
sub deprecated_reserve_PBS_nodes_mainleader($$$){
    my $finalhosts = shift;
    my $mainleader = shift;
    my $wdir = shift;

    my $nNodes = 0;
    my @finalhostfiles = @{$finalhosts};

    my $qsubloc = "/usr/local/bin";
    my $qsubstr = "$qsubloc/qsub -l nodes=";

    my @info = ();

    for(my $i = $#finalhostfiles-1 ; $i>=0 ; $i--){
        @info = split /\./, $finalhostfiles[$i];

        if($finalhostfiles[$i] =~ /pbs/){
            open PBSNODES, "< $wdir/$finalhostfiles[$i]";
            my @nodes = <PBSNODES>;
            $nNodes += ($#nodes+1);
        }
    }

    my $qsubstr2 = $qsubstr.($nNodes);
    my $txferfilename = "pvmaddatonce.pl";

    my $rm_pvmaddcmd = "ssh $mainleader 'rm ~/$txferfilename'";
    `$rm_pvmaddcmd`;
    my $txfer_pvmaddcmd = "cat $txferfilename | ssh $mainleader 'cat - > ~/$txferfilename'";#	`$txfer_pvmaddcmd`;
    my $chmod_pvmaddcmd = "ssh $mainleader 'chmod 700 ~/$txferfilename'";
    `$chmod_pvmaddcmd`;
    my $add2pvm = "ssh $mainleader 'echo \"perl ~/$txferfilename $mainleader $nNodes\" | $qsubstr2'";#	`$add2pvm`;

    print "\nreserving nodes by PBS scheduler in '$info[1]'...\n";
}

#
# reserve pbs-scheduled cluster nodes using "pbsdsh" (deprecated)
#
sub deprecated_reserve_PBS_nodes_pbsdsh($$$){
    my $finalhosts = shift;
    my $mainleader = shift;
    my $wdir = shift;

    my @finalhostfiles = @{$finalhosts};
    for(my $i = $#finalhostfiles-1 ; $i>=0 ; $i--){
        my @info = split /\./, $finalhostfiles[$i];

        print "\nreserving nodes on '$info[1]'...\n";

        if($finalhostfiles[$i] =~ /pbs/){
            my $qsubloc = "/usr/local/bin";
            my $qsubstr = "$qsubloc/qsub -l nodes=";

            open PBSNODES, "< $wdir/$finalhostfiles[$i]";
            foreach(<PBSNODES>){
                chomp $_;
                $qsubstr .= "$_+";
            }
            $qsubstr = substr $qsubstr, 0, length($qsubstr)-1 ;
            close PBSNODES;

            # actual reservation
            my $rm_pvmaddcmd = "ssh $mainleader 'rm ~/pvmadd.sh'";
            `$rm_pvmaddcmd`;
            my $txfer_pvmaddcmd = "cat pvmadd.sh | ssh $mainleader 'cat - > ~/pvmadd.sh'";
            #`$txfer_pvmaddcmd`;
            my $chmod_pvmaddcmd = "ssh $mainleader 'chmod 700 ~/pvmadd.sh'";
            `$chmod_pvmaddcmd`;
            my $add2pvm = "ssh $mainleader 'echo \"pbsdsh ~/pvmadd.sh $mainleader\" | $qsubstr'";
            #`$add2pvm`;
        }
    }
}







#
# BELOW, ALL DEPRECATED FUNCTIONS (NO MORE MAINTAINED...)
#
# Reserve PBS-scheduled nodes by submitting a job with "pvmaddatonce.pl" script
#
# Two variables "clusterheads" and "nNodes" are inout type.
# These variables are modified inside this function "computeNodePerHead"
#
sub deprecated_reserve_PBS_nodes {
    my $self = shift;
    my $FinalMFileSet = shift;
    my $serverDir     = shift;
    my $PVM_startnode = shift;
    my $jobid         = shift;
    my $clusterheads  = shift;
    my $nNodes        = shift;

    my $PVM_startnode_defined = 0;
    my $HPCALE_server_alloc_dir = "$ENV{HPCALE_SERVER_HOME}/server/allocation";
    my $pvmjoin_scr = "pvmadd_script.pl";

    if($PVM_startnode ne "_undefined_"){
        $PVM_startnode_defined = 1;
    }

    $self->computeNodePerHead($FinalMFileSet, $serverDir, $clusterheads, $nNodes);


    for(my $i=0; $i < scalar(@{$clusterheads}) ; $i++){
        my $head = $clusterheads->[$i];
        my $nNode= $nNodes->[$i];

        my ($headhome,
            $headscr,
            $qsub,
            $headjobdir) = $self->get_Locations($head, $jobid, $pvmjoin_scr);

        my $qsubstr = "$qsub -l nodes=".$nNodes->[$i];

        # Copy $pvmjoin_scr to the clusterhead
        `cat "$HPCALE_server_alloc_dir/$pvmjoin_scr" | ssh $head 'cat - >! $headscr '`;
        `ssh $head 'chmod 700 $headscr '`;


        # Allocate by PBS Scheduler of clusterhead
        `ssh $head 'echo \"perl $headscr $PVM_startnode $nNode $PVM_startnode_defined $headjobdir\" | $qsubstr '`;



        my $PVM_sNodeFile = "$headjobdir/pvmstartnode";

        $PVM_startnode = $self->WaitPVM_sNode ($PVM_startnode_defined,
                                               $head,
                                               $PVM_sNodeFile,
                                               $PVM_startnode );
    }

    return $PVM_startnode;
}


#
# Reserve SSH-able nodes and Run PVM on those nodes.
#
sub deprecated_reserve_NonPBS_nodes {
    my $self = shift;
    my $SSH_PVMfile = shift;
    my $serverDir   = shift;
    my $jobid       = shift;

    # Decide PVM Start node. (The first node in the PVMHosts file
    my $PVM_startnode = `head -1 $serverDir/$SSH_PVMfile`;
    chomp $PVM_startnode;

    # Get home directory of PVM Start node.
    my $dirmanager = new RemoteDirManager();
    my $remotehome = $dirmanager->getHomeDir($PVM_startnode);

    # aux. variable (alias)
    my $remote_SSH_PVMfile = "$remotehome/HPCALE_Jobs/$jobid/$SSH_PVMfile";

    my @test = `ls $serverDir/$SSH_PVMfile `;
    print O1 @test;

    # Copy PVMHosts file to PVM Start node
    `cat $serverDir/$SSH_PVMfile | ssh $PVM_startnode 'cat - > $remote_SSH_PVMfile '`;

    # Run PVM
    `ssh $PVM_startnode 'pvm $remote_SSH_PVMfile > /dev/null &'`;

    # Give time to pvm for starting at least one ssh-able node
    system("sleep 1");

    return $PVM_startnode;
}

sub deprecated_get_Locations {
    my $self = shift;
    my $head        = shift;
    my $jobid       = shift;
    my $pvmjoin_scr = shift;

    my $dirmanager = new RemoteDirManager();

    # Get home directory of clusterhead
    my $headhome   = $dirmanager->getHomeDir($head);
    my $headjobdir = "$headhome/HPCALE_Jobs/$jobid";
    my $headscr    = "$headjobdir/$pvmjoin_scr";


    # Get qsub directory of clusterhead
    my $qsub = $dirmanager->checkProgramDir($head, "qsub");

    return ($headhome, $headscr, $qsub, $headjobdir);
}

# compute how many nodes should be allocated for each
# clusterhead. If multiple programs are to be run on same head,
# the number of nodes for the head will be summed up.
#
sub deprecated_computeNodePerHead {
    my $self = shift;
    my $FinalMFileSet = shift;
    my $serverDir     = shift;
    my $clusterheads  = shift;
    my $nNodes        = shift;

    # aux. var
    my @info = ();

    for(my $i = scalar(@{$FinalMFileSet}) - 2 ; $i>=0 ; $i--){
        @info = split /\./, $FinalMFileSet->[$i];

        if($FinalMFileSet->[$i] =~ /pbs/){
            my $file = "$serverDir/".$FinalMFileSet->[$i];

            open PBSNODES, "< $file";
            my @nodes = <PBSNODES>;
            close PBSNODES;

            # the first line of .clst.pbs.hosts is the cluster head
            chomp $nodes[0];

            # To prevent duplicate leaders
            my $idx = -1;
            for(my $j=0 ; $j < scalar(@{$clusterheads}) ; $j++){
                if($nodes[0] eq $clusterheads->[$j]){
                    $idx = $j;
                }
            }

            if($idx == -1){
                push @{$clusterheads}, $nodes[0];
                push @{$nNodes}, ($#nodes+1);
            }
            else{
                $nNodes->[$idx] += ($#nodes+1);
            }

        }
    }
}

# This is kind of lock
# When we try to launch pvm at a PBS cluster, it takes time.
# So, this part waits until
sub deprecated_WaitPVM_sNode {
    my $self = shift;
    my $PVM_sNode_defined = shift;
    my $head = shift;
    my $PVM_sNodeFile = shift;
    my $PVM_startnode = shift;

    my $tempFile = $PVM_sNodeFile."_cat";
    my $fexist = 0;


    if($PVM_sNode_defined == 0){
        do {
            # if $PVM_sNodeFile is created by the pvmjoin_scr,
            # copy it to $PVM_sNodeFile_cat.
            # if $PVM_sNodeFile is not created yet, cat will generate
            # error message and be sinked to /dev/null
            my $cmd = "ssh $head '(cat $PVM_sNodeFile >! $tempFile ) > & /dev/null'";
            `$cmd`;

            # cat the content of $PVM_sNodeFile_cat
            my $aNode = `ssh $head 'cat $tempFile '`;
            if ($aNode =~ /<(.*)>/){
                $fexist = 1;
                $PVM_startnode = $1;
            }
            else{
                `sleep 2`;
            }
        } while($fexist == 0);

        $PVM_sNode_defined = 1;
    }

    return $PVM_startnode;
}




1;

