# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package ExecuteMPIComp;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}/server");
}

use util::Debug;
use util::RemoteDirManager;
use db::PerlDB;
use strict;

my $mod = "ExecuteMPIComp";
my $pipe = "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";

sub new {
    my $classname  = shift;         # What class are we constructing?
    my $self   = {};                # Allocate new memory

    bless($self, $classname);       # Mark it of the right type

    return $self;
}

sub collect_basic_info
{
   my ($self, $comp, $key, $jobid, $jobdir) = @_;

   my $type = $comp->{type};
   my $clst = $comp->{cluster};
   my $head = $comp->{head};
   my $path = $comp->{path};
   my $file = $comp->{file};
   my $nNode= $comp->{nNode};

   my $MFile = "MF-".$key."-".$comp->{name}."-".$comp->{rcid};

   $head =~ /[a-zA-Z0-9]+\.(.*)/;       
   my $domain = $1;

   # collect basic info for launching a component
   my $launchnode = `head -1 $jobdir/$MFile`; chomp $launchnode;
   my $rhome = (new RemoteDirManager())->getHomeDir($launchnode);
   my $rdir  = "$ENV{HPCALE_REMOTE_JOBDIR}/$jobid";

   return ($type, $clst, $head, $path, $file, $nNode, 
           $MFile, $domain,
           $launchnode, $rhome, $rdir);
}

sub get_mpi_bin
{
   my ($lprog, $head) = @_;
   Debug::dbgmsg(3, $mod, $pipe, "Get mpibin progpath and name");

   if(length($lprog) > 0){
      return $lprog;
   }
   else{
      return "mpirun";
   }
}

sub get_mpi_arg
{
   my ($larg, $rhome, $rdir, $MFile, $nNode) = @_;
   my $mpiarg = "";

   if(length($larg) > 0){
      # this case means that user will completely specify the MPI arguments
      $mpiarg = $larg;
      if($mpiarg =~ /__MFILE__/){
         $mpiarg =~ s/__MFILE__/$rhome\/$rdir\/$MFile/g;
      }
   }
   else{
      # apply default MPI arguments
      $mpiarg .= "-np $nNode -nodes $nNode ";
      $mpiarg .= "-machinefile $rhome/$rdir/$MFile ";
   }

   return $mpiarg;
}

#
# <run example>
# mpirun -np 8 -nodes 8 -machinefile xxxx executable 
#
sub executeComponent 
{
   my ($self, $comp, $key, $jobid, $jobdir, $xjdprefix) = @_;

   my ($type, $clst, $head, $path, $file, $nNode,  
       $MFile, $domain, 
       $launchnode, $rhome, $rdir) = $self->collect_basic_info($comp,
                                                               $key, 
                                                               $jobid,
                                                               $jobdir);

   # if machine file is not generated yet, generate here. 
   # Sometimes, machine files is generated at runtime environment setup phase. 
   # (e.g. PVM application)
   $self->buildMachinefiles($comp, $jobdir, $MFile, $domain);
   
   # write machine file and XJD file on remote resource 
   `cat $jobdir/$MFile | ssh $head 'cat - >! $rhome/$rdir/$MFile '`;
   `cat $jobdir/$xjdprefix.xjd | ssh $head 'cat - >! $rhome/$rdir/$xjdprefix.xjd '`;

   Debug::dbgmsg(3, $mod, $pipe, "$type,\n $clst,\n $head,\n $path,\n $file,\n $nNode,\n $MFile,\n $domain,\n $launchnode,\n $rhome,\n $rdir");
   
   # generate command to launch component
   Debug::logmsg($mod, $pipe, "Generating execution command");

   my $prefix = "ssh $launchnode '/bin/sh -c \" "; 
   my $postfix= " \"' ";

   # cd to remote home directory
   my $cd = "cd $rhome/$rdir"; 

   # set MPI bin and MPI-specific arguments
   my $mpibin = &get_mpi_bin($comp->{launchprog}, $head); 
   if(not defined $mpibin){
      return -1;
   }
   my $mpiarg = &get_mpi_arg($comp->{launcharg}, $rhome, $rdir, $MFile, $nNode); 

   # set program path, name and arguments 
   my $prog = "$path/$file"; # set program path and name
   my $progarg = "";
   if(defined $comp->{argument}){
      $progarg = " $comp->{argument} ";
      $progarg =~ s/__MFILE__/$rhome\/$rdir\/$MFile/g;
      $progarg =~ s/__RHOME__/$rhome\/$rdir/g;
   }

   my $xjd = "$rhome/$rdir/$xjdprefix.xjd";

   # set program output pipes
   my $stdout = "$rhome/$rdir/$key-STDOUT.log";
   my $stderr = "$rhome/$rdir/$key-STDERR.log";

   Debug::dbgmsg(3, $mod, $pipe, "[$prefix]");
   Debug::dbgmsg(3, $mod, $pipe, "[$cd]");
   Debug::dbgmsg(3, $mod, $pipe, "[$mpibin]");
   Debug::dbgmsg(3, $mod, $pipe, "[$mpiarg]");
   Debug::dbgmsg(3, $mod, $pipe, "[$prog]");
   Debug::dbgmsg(3, $mod, $pipe, "[$progarg]");
   Debug::dbgmsg(3, $mod, $pipe, "[$stdout]");
   
   my $command = "$prefix $cd; $mpibin $mpiarg $prog $progarg >$stdout 2>$stderr $postfix";

   Debug::logmsg($mod, $pipe, "[$command]");

   # background execution and return
   my $v = system "$command &";
   ($v == 0) ? return 1 : return -1; 
}

sub buildMachinefiles 
{
   my ($self, $comp, $jobdir, $MFile, $domain) = @_;
   my $result = 1;

   if(-e "$jobdir/$MFile"){
      Debug::logmsg($mod, $pipe, "Machinefile $MFile already exist.");
      return 1;
   }
    
   if(defined $comp->{cluster}){
      if(defined $comp->{nodeset}){
         open MF, "> $jobdir/$MFile" or return -1;
         my @nodes = split /,/ , $comp->{nodeset} ;        
         foreach(@nodes){
            chomp $_;
            print MF "$_.$domain\n";
         } 
         close MF;
      }
      else{
         $result = $self->genClusterMachineFile($jobdir, $MFile, $comp);          
      }                        
   }
   else{
      open MF, "> $jobdir/$MFile" or return -1;
      print MF "$comp->{node}.$domain\n";
      close MF;
   }      

   return $result;
}

sub genClusterMachineFile 
{
   my ($self, $jobdir, $MFile, $comp) = @_;
    
   my $head = $comp->{head};
   $head =~ /[a-zA-Z0-9]+\.(.*)/;       
   my $domain = $1;
    
   if($comp->{scheduler} eq "PBS"){
      open PBSNODE, "< $jobdir/PBSNODES-$head" or return -1;
   }
   my @nodes = <PBSNODE>;
   close PBSNODE;
    
    open MF, "> $jobdir/$MFile";
    open TMP, "> $jobdir/tmp";

    my $i = 0;            
    foreach(@nodes){
        chomp $_;
        if($i < $comp->{nNode}){
            print MF "$_.$domain\n";
            $i++;
        }
        else{
            print TMP "$_\n";
        }
    }
    close MF;
    close TMP;
    
    `mv $jobdir/tmp $jobdir/PBSNODES-$head`;
    
    return 1;
}

1;
