# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package JobExecManager;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}");
    push (@INC, "$ENV{HPCALE_SERVER_HOME}/server");
}

use util::Debug;
use util::RemoteDirManager;
use util::JobRecordManager;
use execution::ExecuteMPIComp;
use execution::ExecutePVMComp;
use execution::ExecuteSEQComp;
use strict;

my $mod = "JobExecManager";
my $pipe= "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";

sub new 
{
   my $classname  = shift;         # What class are we constructing?
   my $self   = {};                # Allocate new memory

   bless($self, $classname);       # Mark it of the right type
   return $self;
}

sub launch_comp 
{
   my ($self, $comp, $key, $jobid, $jobdir, $xjdprefix) = @_;
   Debug::logmsg($mod, $pipe, "Launching $comp($key)...");

   my $result = undef;
   if($comp->{type} eq "MPI-executable"){
      Debug::dbgmsg(2, $mod, $pipe, "$comp($key) is MPI-executable");
      $result = (new ExecuteMPIComp())->executeComponent($comp,
                                                         $key,
                                                         $jobid,
                                                         $jobdir,
                                                         $xjdprefix
                                                        );
   }
   elsif($comp->{type} eq "PVM-executable"){
      Debug::dbgmsg(2, $mod, $pipe, "$comp($key) is PVM-executable");
      $result = (new ExecutePVMComp())->executeComponent($comp,
                                                         $key,
                                                         $jobid,
                                                         $jobdir,
                                                         $xjdprefix
                                                        );
   }
   elsif($comp->{type} eq "Executable"){
      Debug::dbgmsg(2, $mod, $pipe, "$comp($key) is Executable");
      $result = (new ExecuteSEQComp())->executeComponent($comp,
                                                         $key,
                                                         $jobid,
                                                         $jobdir,
                                                         $xjdprefix
                                                        );
   }

   return $result;
}


#
# kill a job explicitly.
# the structure of this subroutine is same to the checkJobs subroutine.
#
sub stopExecution 
{
   my ($self, $jobdir, $runningComp, $jobRecordMgr, $jobid) = @_;         
    
   my @nodes = ();
   foreach(@{$runningComp}){
      my @rcomp = @{$_};

      open I, "< $jobdir/$rcomp[0]"; #open machine file
      foreach(<I>){
         chomp $_;
         
         my $ostype = $self->checkOSType($_);

         push @nodes, [$_, $rcomp[1], $ostype]; 
         Debug::dbgmsg(3, $mod, $pipe, "ToStop: $_: $rcomp[1], $rcomp[0], $ostype");
      }
      close I;
   }

   # loop the nodes and kill the job
   my $canEnd = 0;
   while($canEnd == 0){
      $canEnd = 1;
      my @reduced_nodes = ();
      foreach my $i (0 .. $#nodes){
         my @ncomp = @{$nodes[$i]};
         my @result = $self->checkProcess($nodes[$i]);
           
         # assume that the first line of output is the title of output "PID"
         if($#result < 0){ 
            Debug::dbgmsg(3, $mod, $pipe, "Component $ncomp[1] is not running at $ncomp[0]");
         }
         else{
            my $running = 0;
            foreach(@result){
               my @t = split /\s+/, $_;
               if($t[1] =~ /$ncomp[1]/){
                  $running = 1;
                  Debug::dbgmsg(3, $mod, $pipe, "Component $ncomp[1] is running at $ncomp[0] with PID $t[0]");

                  # kill with pid
                  `ssh $ncomp[0] '/bin/sh -c \"kill -9 $t[0]\"'`;

                  # save to check again
                  push @reduced_nodes, $nodes[$i];
               }
            }
            $canEnd = 0 if($running == 1);
         }
      }

      @nodes = @reduced_nodes;
      @reduced_nodes = ();
   }

   # if arrives here, assume that all jobs are terminated. 
   $jobRecordMgr->update_record($jobid, 2, "expelled");
}        


sub checkOSType 
{
   my ($self, $host) = @_;
   my $ostype = `ssh $host 'uname'`;
   chomp $ostype;

   return $ostype;
}

#
# return set of process information in array
# act differently according to the uname
#
sub checkProcess
{
   my ($self, $nodes) = @_;
   my @ncomp = @{$nodes};
   
   my $host = $ncomp[0];
   my $comm = $ncomp[1];
   my $ostype = $ncomp[2];
   
   # previously used commands
   #my $cmd = "ssh $ncomp[0] 'ps -elf | grep $ncomp[1] '";
   #my $cmd = "ssh $ncomp[0] 'ps -C $ncomp[1] -o \"pid\" '";
   #my $cmd = "ssh $ncomp[0] '/bin/sh -c \"pgrep -lf $ncomp[1]\"'";

   my $cmd = "";
   if ($ostype eq "AIX"){
      $cmd = "ssh $host '/bin/sh -c \"ps -e -o pid,comm | grep $comm\"'";
   }
   elsif($ostype eq "Linux"){
      $cmd = "ssh $host '/bin/sh -c \"ps -e -o pid,comm | grep $comm\"'";
   }
   elsif($ostype eq "SunOS"){
      $cmd = "ssh $host '/bin/sh -c \"ps -e -o pid,comm | grep $comm\"'";
   }

   my (@result) = `$cmd`;
   return @result;
}


#
# check whether given job is running on the reserved nodes.
# We can figure out that by polling each host using 'ps' command
#
sub checkJobs 
{
   my ($self, $jobdir, $runningComp, $jobRecordMgr, $jobid) = @_;         
   Debug::logmsg($mod, $pipe, "checking job termination...");

   my @nodes = ();
   foreach(@{$runningComp}){
      my @rcomp = @{$_};

      open I, "< $jobdir/$rcomp[0]"; #open machine file
      foreach(<I>){
         chomp $_;
         my $ostype = $self->checkOSType($_);

         push @nodes, [$_, $rcomp[1], $ostype];
         Debug::dbgmsg(3, $mod, $pipe, "$_: $rcomp[1], $rcomp[0], $ostype");
      }
      close I;
   }
   
   my $canEnd = 0;
   while($canEnd == 0){
      $canEnd = 1;
      foreach my $i (0 .. $#nodes){
         my @ncomp = @{$nodes[$i]};
         my @result = $self->checkProcess($nodes[$i]); 
         
         # assume that the first line of output is the title of output "PID"
         if($#result < 0){ 
            Debug::dbgmsg(3, $mod, $pipe, "Component $ncomp[1] is not running at $ncomp[0]. (May finished already?)");
         }
         else{
            my $running = 0;
            foreach(@result){
               my @t = split /\s+/, $_;
               if($t[1] =~ /$ncomp[1]/){
                  $running = 1;
                  Debug::dbgmsg(3, $mod, $pipe, "Component $ncomp[1] is running at $ncomp[0] with PID $t[0]");
               }
            }
            $canEnd = 0 if($running == 1);
         }
      }
   }

   # if arrives here, assume that all jobs are finished
   $jobRecordMgr->update_record($jobid, 2, "terminated");
}

1;


