# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package RTEnvAnalyze;

BEGIN{
    push (@INC, "$ENV{HPCALE_SERVER_HOME}");
    push (@INC, "$ENV{HPCALE_SERVER_HOME}/server");
}

use db::PerlDB;
use util::Debug;
use util::XJDParser;
use runtimeenv::PVMHandler;
use runtimeenv::FileTransferHandler;
use Data::Dumper;
use strict;

my $mod = "RuntimeEnv";
my $pipe= "$ENV{HPCALE_LOGPATH}/$ENV{HPCALE_LOGFILE}";

sub new {
   my $classname  = shift;         # What class are we constructing?
   my $self   = {};                # Allocate new memory

   bless($self, $classname);       # Mark it of the right type

   $self->{RThandler} = ();
   $self->{chost} = undef;
   $self->{user} = undef;
   $self->{jobdir} = undef;
    
   return $self;
}

sub reorganize
{
   my ($self, $user, $chost, $jobid, $lowXJD) = @_;

   my $jobdir = "$ENV{HPCALE_JOBDIR}/$jobid";
   $self->{user} = $user;
   $self->{chost} = $chost;
   $self->{jobdir} = $jobdir;

   my $compList = $lowXJD->getComponentList();
   foreach my $key (keys %{$compList}){
      my $comp = $compList->{$key};

      my ($head,$name,$cid) = &getValues($comp);
      my @requirements = &getRuntimeReq($cid);               

      # File transfer handler should be registered unconditionally
      # due to the stdout and stderr.        
      push @requirements, "FTP";

      foreach my $req (@requirements){
         my $handler = undef;
        
         if($req eq "InterComm"){
            Debug::dbgmsg(3, $mod, $pipe, "PVM (InterComm) required for $name");
            $handler = $self->get_pvm_handler($jobid, $jobdir);
         }
         elsif($req eq "PVM"){
            Debug::dbgmsg(3, $mod, $pipe, "PVM Handler required for $name");
            $handler = $self->get_pvm_handler($jobid, $jobdir);
         }
         elsif($req eq "FTP"){
            Debug::dbgmsg(3, $mod, $pipe, "FileTransfer Handler required for $name");
            $handler = $self->get_ftp_handler($jobid,$jobdir,$chost,$user);
         }
         else{
            Debug::logmsg($mod, $pipe, "Unsupported Requirements $req for $name");
            return -1;
         }

         return -1 if ($self->handler_reorganize($handler,$comp,$key,$name,$head) == -1);
      }
   }

   Debug::logmsg($mod, $pipe, "Runtime environment info reorganization: ok!! ");
   return 1;
}

sub prepare
{
   my ($self, $user, $chost, $jobid, $lowXJD) = @_;

   my $jobdir = "$ENV{HPCALE_JOBDIR}/$jobid";
   $self->{user} = $user;
   $self->{chost} = $chost;
   $self->{jobdir} = $jobdir;
    
   my $compList = $lowXJD->getComponentList();
   foreach my $key (keys %{$compList}){
      my $comp = $compList->{$key};

      my ($head,$name,$cid) = &getValues($comp);
      my @requirements = &getRuntimeReq($cid);               

      # File transfer handler should be registered unconditionally
      # due to the stdout and stderr.        
      push @requirements, "FTP";
        
      Debug::dbgmsg(3, $mod, $pipe, "Req: @requirements"); 

      foreach my $req (@requirements){
         my $handler = undef;
        
         if($req eq "InterComm"){
            Debug::dbgmsg(3, $mod, $pipe, "PVM (InterComm) required for $name");
            $handler = $self->get_pvm_handler($jobid, $jobdir);
         }
         elsif($req eq "PVM"){
            Debug::dbgmsg(3, $mod, $pipe, "PVM Handler required for $name");
            $handler = $self->get_pvm_handler($jobid, $jobdir);
         }
         elsif($req eq "FTP"){
            Debug::dbgmsg(3, $mod, $pipe, "FileTransfer Handler required for $name");
            $handler = $self->get_ftp_handler($jobid,$jobdir,$chost,$user);
         }
         else{
            Debug::logmsg($mod, $pipe, "Unsupported Requirements $req for $name");
            return -1;
         }

         return -1 if ($self->handler_prepare($handler,$comp,$key,$name,$head) == -1);
      }
   }

   return 1;
}

sub beReady
{
   my ($self, $notifier) = @_;
   Debug::logmsg($mod, $pipe, "[Making Runtime Environments be ready...]");

   foreach (keys %{$self->{RThandler}}){
      Debug::logmsg($mod, $pipe, "[Handler: $_]");
        
      my $r = undef;
      if($_ eq "FTP"){
         $r = $self->{RThandler}->{$_}->beReady($notifier) ;
      }
      elsif($_ eq "PVM"){
         $r = $self->{RThandler}->{$_}->beReady ;
      }

      if($r != 1){
         Debug::errmsg($mod, $pipe, "[Problem making runtime requirements ready]");
         return -1;
      }
   }

   Debug::logmsg($mod, $pipe, "[All requirements are ready now..]");
   return 1;
}

sub back_transfer{
   # the notifier is ExecutionNotifier in this subroutine.
   my ($self,$notifier) = @_;

   if(defined $self->{RThandler}->{FTP}){
      my $v = $self->{RThandler}->{FTP}->back_transfer($notifier) ;
      return -1 if ($v < 0);
   }

   return 1;
}

sub cleanupRTenv {
   my $self = shift;

   foreach (keys %{$self->{RThandler}}){
      Debug::logmsg($mod, $pipe, "Cleaning $_ handler");
      $self->{RThandler}->{$_}->cleanup ;
   }
}

sub get_pvm_handler
{
   my ($self, $jobid, $jobdir) = @_;

   if(!defined($self->{RThandler}->{PVM})){
      $self->{RThandler}->{PVM} = new PVMHandler($jobid, $jobdir);
   }
   else{
      Debug::logmsg($mod, $pipe, "PVM handler already exists.");
   }

   return $self->{RThandler}->{PVM};
}

sub get_ftp_handler
{
   my ($self, $jobid, $jobdir, $chost, $user) = @_;

   if(!defined($self->{RThandler}->{FTP})){
      $self->{RThandler}->{FTP} = new FileTransferHandler($jobid, $jobdir, 
                                                          $chost, $user);
   }
   else{
      Debug::logmsg($mod, $pipe, "FTP Handler is already registered");
   }
           
   return $self->{RThandler}->{FTP};
}

sub handler_prepare
{
   my ($self, $handler, $comp, $key, $name, $head) = @_;

   if($handler->prepare($comp, $key) < 0){
      $handler->error ($name, $head) ;
      Debug::dbgmsg(3, $mod, $pipe, "prepare $handler->{name} Handler: fail");
      return -1;
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "prepare $handler->{name} Handler: ok");
      return 1;
   }
}

sub handler_reorganize
{
   my ($self, $handler, $comp, $key, $name, $head) = @_;

   if($handler->reorganize($comp, $key) < 0){
      $handler->error ($name, $head) ;
      Debug::dbgmsg(3, $mod, $pipe, "reorganize $handler->{name} Handler: fail");
      return -1;
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "reorganize $handler->{name} Handler: ok");
      return 1;
   }
}

sub getValues {
   my $comp = shift;
   my $head = $comp->{head};
   my $name = $comp->{name}; # actually, virtual name of the component
   my $rc_id = $comp->{rcid}; # means location of the component 
   my $cid = &getCID($name, $rc_id) ;

   Debug::dbgmsg(3, $mod, $pipe, "CompName: $name, Head: $head, RCID: $rc_id");
   Debug::dbgmsg(3, $mod, $pipe, "Component ID: $cid");
   
   return ($head,$name,$cid);
}


sub getCID {
    my $name = shift;
    my $rc_id = shift;

    my $perldb = new PerlDB();

    my $query = "select id from component where compID = \"$name\" and rc_id = $rc_id";
    my $cid = $perldb->select($query);
    
    $perldb->db_end ();

    return $cid->[0];
}

sub getRuntimeReq {
    my $name = shift;
    my $perldb = new PerlDB();
    my @reqs = ();

    my $query = "select req_name from runtimereq where component_id = $name";
    my @result = $perldb->select($query);
    $perldb->db_end ();

    foreach(@{$result[0]}){
        push @reqs, $_;
    }
    
    if(scalar(@reqs) == 0){
        return ();
    }
    else{
        return @reqs;
    }
}

1;
